using Gtk;

public class CategoriesView : Gtk.HPaned {

    private TasksTreeView tasksView;
	private CategoriesTreeView categoriesView;
	private Gtk.TreeModelFilter filter;
	private Gtk.ListStore categoriesModel;
	private IBackend backend;
	private Gtk.TreeIter selectedIter;
    private Gtk.Button deleteCategoryButton;

	private bool showDone;

	private string searchString;
	private Logger log;

	construct {
		searchString = "";
		log = Logger.getLogger("CategoriesView");
		showDone = false;
	}

    public CategoriesView (IBackend aBackend, Gtk.Widget anEditArea)
    {
		backend = aBackend;

		categoriesView = new CategoriesTreeView (true);
		categoriesView.currentFilterRule = categoriesView.ALL_CATEGORIES;
		categoriesView.nameRenderer.edited += editCategoryName;
	
		Gtk.HBox toolbar = new Gtk.HBox(true,0);
		Gtk.Button addCategoryButton = new Gtk.Button.from_stock(Gtk.STOCK_ADD);
		deleteCategoryButton = new Gtk.Button.from_stock(Gtk.STOCK_REMOVE);

	    int size_not_deleted = 0;
	    foreach (ICategory category in this.backend.categories) {
	        if (!category.isDeleted) {
	            size_not_deleted ++;
	        }
	    }
        if (size_not_deleted < 2) deleteCategoryButton.sensitive = false;
        
		addCategoryButton.clicked += () => {
			ICategory category = backend.newCategory(true);
			category.created += (sender) => {
				this.update();
			};
			category.push();
			deleteCategoryButton.sensitive = true;
		};

		deleteCategoryButton.clicked += deleteCategory;

		toolbar.add(addCategoryButton);
		toolbar.add(deleteCategoryButton);

		Gtk.VBox optionsBox = new Gtk.VBox(false, 0);
		Gtk.CheckButton doneOption = new Gtk.CheckButton.with_label (_("Show completed task"));
		doneOption.toggled += () => { 
			this.showDone = (this.showDone) ? false : true; 
			this.filter.refilter();
			};
		optionsBox.add( doneOption);

		Gtk.VBox categoriesBox = new Gtk.VBox(false, 0);
		categoriesBox.pack_start(categoriesView, true, true, 0);
		categoriesBox.pack_start(optionsBox, false, false, 0);
		categoriesBox.pack_start(toolbar, false, false, 0);


		tasksView = new TasksTreeView (anEditArea, false);

        Adjustment hScrollAdjustment = new Adjustment (0,0,0,0,0,0);
        Adjustment vScrollAdjustment = new Adjustment (0,0,0,0,0,0);
        ScrolledWindow scrolledWindow = new ScrolledWindow (hScrollAdjustment, vScrollAdjustment);
        scrolledWindow.hscrollbar_policy = PolicyType.NEVER;
        scrolledWindow.vscrollbar_policy = PolicyType.AUTOMATIC;

        scrolledWindow.add(tasksView);

		position = 180;
		pack1 (categoriesBox, false, false);
		pack2 (scrolledWindow, true, false);
    }

	public void update() {
			updateTasksModel();
			this.tasksView.set_model ( this.filter );
			categoriesView.targetModel = this.filter;
			this.filter.refilter();

			updateCategoriesModel();
			this.categoriesView.set_model ( this.categoriesModel );
			Gtk.TreePath? path = this.categoriesModel.get_path (this.selectedIter);
			if (path != null) this.categoriesView.set_cursor(path, null, false);

	}

	public void filterSearch(string text) {
		this.searchString = text;
		//TODO add timeout for search
		this.filter.refilter();
	}

/******************************************************************************
 * PRIVATE
 */

	private void editCategoryName ( GLib.Object sender, string path, string newName) {
        Gtk.TreeIter iter;
        ICategory category;
        
        Gtk.TreePath aPath = new Gtk.TreePath.from_string( path );
        if (!this.categoriesModel.get_iter (out iter, aPath))
	    return; // Do nothing
				
		this.categoriesModel.get (iter, 0, out category);	
        if ( category == null ) return;		
		if (category.name != newName) {
            category.name = newName;
            category.update();       
        }
    }


	private void updateTasksModel() {
        Gtk.TreeIter ti;
        Gtk.ListStore treeModel = new Gtk.ListStore (1, typeof(ITask));

        /*
         unpack the task list and store it in a tree model
        */
        foreach (ITask task in this.backend.tasks) {
            treeModel.append (out ti);
            treeModel.set (ti, 0, task);
        }

 		filter = new Gtk.TreeModelFilter (treeModel, null);
		filter.set_visible_func( filterTasksTree );
	}

	private void updateCategoriesModel() {

        Gtk.TreeIter ti;
		CategoriesListStore treeModel = new CategoriesListStore();

        /*
         unpack the task list and store it in a tree model
        */
		ICategory meta;
		meta = backend.newCategory(false);
		meta.name = "<b>" + _("All Categories") + "</b>";
		meta.id = categoriesView.ALL_CATEGORIES;
        treeModel.append (out ti);
        treeModel.set (ti, 0, meta);
		selectedIter = ti;

		Utils.Backend.sortCategoriesByPositionAsc(this.backend.categories);
        foreach (ICategory category in this.backend.categories) {
				if (!category.isDeleted) {
			        treeModel.append (out ti);
			        treeModel.set (ti, 0, category);
				}
        }
		categoriesModel = treeModel;
	}

    private bool filterTasksTree (Gtk.TreeModel model, Gtk.TreeIter iter) {
 	    ITask task;
	    model.get (iter, 0, out task);
		if (task == null) return false;

		if ((task.isDeleted)||(!this.showDone && task.done)) return false;

	    if (
				((categoriesView.currentFilterRule == categoriesView.ALL_CATEGORIES) ||
				(task.categoryID == categoriesView.currentFilterRule) ) &&

				( (searchString == "") || (searchString == Utils.Constant.filterTasks) ||
			  (task.name.casefold().str(searchString.casefold()) != null) || 
			  (task.text.casefold().str(searchString.casefold()) != null) )

			) {
            return true;
        } else {
           return false;
		}
    }
    
    private void deleteCategory (Gtk.Button button) {
		ICategory category = this.categoriesView.getSelectedCategory();
		if (category == null) {
		    log.error("Can not delete 'null' category");
		    return;
		}

		//prevent the deletion of last category
		int size_not_deleted = 0;
		foreach (ICategory a_category in this.backend.categories) {
		    if (!a_category.isDeleted) {
		        size_not_deleted ++;
		    }
		}
		if (size_not_deleted < 2) return;
		
		//disable the delete button
		if (size_not_deleted < 3) deleteCategoryButton.sensitive = false;
		
		Utils.Backend.sortCategoriesByPositionAsc(this.backend.categories);		
		ICategory new_category = null;
        foreach (ICategory a_category in this.backend.categories) {
		    if (!a_category.isDeleted) {
		        if (a_category != category) {
		            new_category = a_category;
		            break;
		        }
		    }
		}
        if (new_category != null) {
            foreach (ITask task in this.backend.tasks) {
                if (task.categoryID == category.id) {
                    task.categoryID = new_category.id;
                }
            }
        }
     
		/* now we can delete the category */
		category.delete();
	}

}

public class CategoriesListStore : Gtk.ListStore, Gtk.TreeDragDest, Gtk.TreeDragSource {
	private Logger log;
	
	construct {
		log = Logger.getLogger("CategoriesListStore");
	}

	public CategoriesListStore () {
		set_column_types (new GLib.Type[] { typeof(ICategory)});
	}

	public bool drag_data_delete (Gtk.TreePath path) {

	    Gtk.TreeIter iter;
        ICategory category;
        
        if (!get_iter (out iter, path)) return false;
		get (iter, 0, out category);
		if (category == null) return false;

		bool val = base.drag_data_delete (path);

		get_iter_first(out iter);
		get (iter, 0, out category);
		int i = 1;
		while (iter_next(ref iter)) {
			get (iter, 0, out category);
			if (category.position != i ) {
				category.position = i;
				category.update();
			}
			i++;
		}
		category.backend.updated();
		return val;
	}

	public bool row_drop_possible (Gtk.TreePath dest, Gtk.SelectionData selection_data) {

	    Gtk.TreeIter iter;
        ICategory category;
        
        if (!get_iter (out iter, dest)) return false;
		get (iter, 0, out category);
		if (category == null) return false;

		if (category.id == CategoriesTreeView.ALL_CATEGORIES) {
			return false;
		} else {
			return true;
		}
	}

}
