using GLib;
using Gtk;

public class MainWindow : Gtk.Window {

    public Gtk.Widget editArea;

    private IBackend backend;
    private Gtk.EventBox contentBox;
    private Gtk.EventBox toolbarBox;
    private Gtk.HBox taskBox;
    private Gtk.HBox noteBox;
    private Gtk.HBox searchBox;
    private Gtk.Statusbar statusbar;
    private Gtk.MenuToolButton viewButton;

    private PreferencesDialog preferencesDialog;
    private Configuration preferences;

	private HomeView homeView;
	private NotesView notesView;
	private ProjectsView projectsView;
	private CategoriesView categoriesView;
	private RecoveryView recoveryView;
	private TicklerView ticklerView;
	private CalendarView calendarView;

	private Gtk.Entry noteSearchEntry;
	private Gtk.Entry taskSearchEntry;
	private Gtk.Entry searchEntry;

	private Logger log;
    private enum viewMode { HOME = 0, PROJECTS, CALENDAR, CATEGORIES, TICKLER , NOTES, SEARCH, RECOVER}
    public int currentViewMode;

	private bool backendUpdated;
	construct {
        this.title = _("Lucruri");
        this.icon = Utils.UI.getPixbuf ("/icons/16x16/emotes/face-monkey.png" );
		notesView = null;
		log = Logger.getLogger("MainWindow");
		backendUpdated = false;
	}

    public MainWindow (IBackend aBackend, Configuration somePreferences) {

        this.backend = aBackend;
        this.preferences = somePreferences;
		currentViewMode = this.preferences.view;
		
        this.backend.notFound += ( sender ) => {
            //%s is the name of the backend
            Utils.UI.showErrorDialog (this, _("%s could not be found at the location specified in the configuration. Please check the preferences window.").printf(backend.name) );
        };

        this.backend.accessDenied += ( sender ) => {
            //%s is the name of the backend
            Utils.UI.showErrorDialog (this, _("Your credentials were not accepted by %s. Please check the prefereces window.").printf(backend.name));
        };

		this.backend.updated.connect(this.updateView);
		
        contentBox = new Gtk.EventBox();
		toolbarBox = new Gtk.EventBox();
        editArea = new Gtk.EventBox();

        initializeView();
        //refresh at every 5 minutes
        //TODO add a configure option
        //GLib.Timeout.add_seconds(5 * 60  , update);
    }

    public bool update () {
        // If backend is not configured show preferences dialog
        if (!this.backend.isConfigured ) {
            // FIXME see why we get segfault on this
            //showInfoDialog (_("%s is not configured. After closing this information window you will be promted to configure %s.").printf(this.backend.name));
            showPreferences( this );
            return false;
        }

        this.backend.update(this.updateView);
        return true;
    }


	public void updateView () {

		if (!backendUpdated) {
			backendUpdated = true;
			preferences.initOpenedWindows();
		}

		//FIXME update category groups
		//if (homeView != null) homeView.updateAll();
		//if (ticklerView != null) ticklerView.updateAll();
		switchView();
	}
	

/******************************************************************************
 * PRIVATE stuff
 */

	private void initializeView () {
        // TODO get size from gconf
		int height = -1;
    	int width = -1;
		if ( width == -1 )	width = 600;
		if ( height == -1 ) height = 600;
		default_width = width;
		default_height = height;

        VBox mainVBox = new VBox(false, 0);
        HBox topHBox = new HBox(false, 0);
 
		/*view button */
        viewButton = new Gtk.MenuToolButton(
	    	new Gtk.Image.from_stock(Gtk.STOCK_HOME,Gtk.IconSize.BUTTON),
	    	_("Home")
	    );
        viewButton.clicked.connect(
            (button) => {
                (button as Gtk.MenuToolButton).menu.popup(
                    null, null,
                    Utils.UI.getMenuPosition,
                    1,
                    Gtk.get_current_event_time ());
            });
        Gtk.Menu mainMenu = createMainMenu();
        viewButton.menu = mainMenu;
	
		/* task stuff */
		taskBox = new Gtk.HBox(false, 0);
		Gtk.Button addTaskButton = new Gtk.Button.with_label(_("Add Task"));
		addTaskButton.set_image(new Gtk.Image.from_stock(Gtk.STOCK_ADD,Gtk.IconSize.BUTTON));
		addTaskButton.clicked += () => { addTask(); };

		Gtk.ToolButton clearTaskSearchButton = new Gtk.ToolButton.from_stock(Gtk.STOCK_CLEAR);
		clearTaskSearchButton.clicked += () => { 
			taskSearchEntry.set_text(Utils.Constant.filterTasks);
		};

		taskSearchEntry = new Utils.Entry(Utils.Constant.filterTasks);
		taskSearchEntry.changed += ( sender ) => {
			string text = (sender as Gtk.Entry).get_text();
			if (this.currentViewMode == this.viewMode.PROJECTS )
				this.projectsView.filterSearch(text);
			else
				this.categoriesView.filterSearch(text);
		};

		taskBox.pack_start (taskSearchEntry, true, true, 0);
		taskBox.pack_start (addTaskButton, false, false, 0);

		/* note stuff */
		noteBox = new Gtk.HBox(false, 0);
		Gtk.Button addNoteButton = new Gtk.Button.with_label(_("Add Note"));
		addNoteButton.set_image(new Gtk.Image.from_stock(Gtk.STOCK_ADD,Gtk.IconSize.BUTTON));
		addNoteButton.clicked += ( button ) => { addNote(); };


		noteSearchEntry = new Utils.Entry(Utils.Constant.filterNotes);
		noteSearchEntry.changed += ( sender ) => {
			string text = (sender as Gtk.Entry).get_text();
			this.notesView.search(text);
		};

		noteBox.pack_start (noteSearchEntry, true, true, 0);
		noteBox.pack_start (addNoteButton, false, false, 0);

		/* search toolbox stuff */
		searchBox = new Gtk.HBox(false, 0);
		Gtk.Button searchAddNoteButton = new Gtk.Button.with_label(_("Add Note"));
		searchAddNoteButton.set_image(new Gtk.Image.from_stock(Gtk.STOCK_ADD,Gtk.IconSize.BUTTON));
		searchAddNoteButton.clicked += ( button ) => { addNote(); };

		Gtk.Button searchAddTaskButton = new Gtk.Button.with_label(_("Add Task"));
		searchAddTaskButton.set_image(new Gtk.Image.from_stock(Gtk.STOCK_ADD,Gtk.IconSize.BUTTON));
		searchAddTaskButton.clicked += ( button ) => { addTask(); };

		searchEntry = new Utils.Entry(Utils.Constant.searchAll);
		searchEntry.icon_press.connect(
		    (sender, position, event) => {
		        if (this.currentViewMode != this.viewMode.RECOVER ) {
			        searchEntry.set_text(Utils.Constant.searchAll);
			        this.homeView.search("");
		        } else {
			        searchEntry.set_text(Utils.Constant.filterRecover);
			        this.recoveryView.filterSearch("");
		        }
	        });
		searchEntry.changed.connect(
		    (sender) => {
		        string text = (sender as Gtk.Entry).get_text();
		        if (text != "") {
			        this.homeView.search(text);
		        }
	        });

		searchBox.pack_start (searchEntry, true, true, 0);
		searchBox.pack_start (searchAddTaskButton, false, false, 0);
		searchBox.pack_start (searchAddNoteButton, false, false, 0);

		/* top stuff */
        topHBox.pack_start (viewButton, false, false, 0);
        topHBox.pack_start (toolbarBox, true, true, 0);

        /*  bottom stuff - for now just statusbar  */
        statusbar = new Gtk.Statusbar ();
		statusbar.has_resize_grip = true;

        /* add main VBox items */
        mainVBox.pack_start (topHBox, false, false, 0);
        mainVBox.pack_start (contentBox , true, true, 0);
        mainVBox.pack_start (editArea , false, false, 0);
        //mainVBox.pack_start (statusbar, false, false, 0);

        add (mainVBox);
    }

    private Gtk.Menu createMainMenu(){
	Gtk.Menu mainMenu = new Gtk.Menu();
        Gtk.SeparatorMenuItem menuSeparator;
        ImageMenuItem menuItem;

        menuItem = new ImageMenuItem.with_label (_("Overview"));
        menuItem.image = new Gtk.Image.from_stock(Gtk.STOCK_HOME, Gtk.IconSize.MENU);
        menuItem.activate += (sender) => {
			this.currentViewMode = this.viewMode.HOME;
			updateView();
		};
		menuItem.always_show_image = true;
        menuItem.show();
        mainMenu.append (menuItem);

        /* Menu entry for notes view.*/
        menuItem = new ImageMenuItem.with_label (_("Notes"));
        menuItem.image = new Gtk.Image.from_pixbuf(Utils.UI.getPixbuf("/icons/16x16/apps/accessories-text-editor.png"));
        menuItem.activate += (sender) => {
	        this.currentViewMode = this.viewMode.NOTES;
			updateView();
		};
		menuItem.always_show_image = true;
        menuItem.show();
        mainMenu.append (menuItem);

        /* Menu entry for tasks view.*/
		menuItem = new ImageMenuItem.with_label (_("Tasks"));
        menuItem.image = new Gtk.Image.from_pixbuf(Utils.UI.getPixbuf("/icons/16x16/actions/bookmark-new.png"));
        menuItem.activate += (sender) => {
	        this.currentViewMode = this.viewMode.CATEGORIES;
			updateView();
		};
		menuItem.always_show_image = true;
        menuItem.show();
        mainMenu.append (menuItem);

		menuSeparator = new Gtk.SeparatorMenuItem();
        menuSeparator.show();
		mainMenu.append (menuSeparator);

        /* Menu entry for projects view.*/
		menuItem = new ImageMenuItem.with_label (_("Projects"));
        menuItem.image = new Gtk.Image.from_pixbuf(Utils.UI.getPixbuf("/icons/16x16/categories/applications-development.png"));
        menuItem.activate += (sender) => {
	        this.currentViewMode = this.viewMode.PROJECTS;
			updateView();
		};
		menuItem.always_show_image = true;
        menuItem.show();
        mainMenu.append (menuItem);

        /* Menu entry for calendar view.*/
        menuItem = new ImageMenuItem.with_label (_("Calendar"));
        menuItem.image = new Gtk.Image.from_pixbuf(Utils.UI.getPixbuf("/icons/16x16/apps/office-calendar.png"));
        menuItem.activate += (sender) => {
	        this.currentViewMode = this.viewMode.CALENDAR;
			updateView();
		};
		menuItem.always_show_image = true;
        menuItem.show();
        mainMenu.append (menuItem);

        /* Menu entry for ticler view.*/	
		menuItem = new ImageMenuItem.with_label (_("Tickler"));
        menuItem.image = new Gtk.Image.from_pixbuf(Utils.UI.getPixbuf("/icons/16x16/status/folder-visiting.png"));
        menuItem.activate += (sender) => {
	        this.currentViewMode = this.viewMode.TICKLER;
			updateView();
		};
		menuItem.always_show_image = true;
        menuItem.show();
        mainMenu.append (menuItem);

		menuSeparator = new Gtk.SeparatorMenuItem();
        menuSeparator.show();	
		mainMenu.append (menuSeparator);

        /* Menu entry for recovering deleted notes and tasks */
        menuItem = new ImageMenuItem.with_label (_("Recover notes and tasks"));
        menuItem.image = new Gtk.Image.from_stock(Gtk.STOCK_UNDO, Gtk.IconSize.MENU);
        menuItem.activate += (sender) => {
	        this.currentViewMode = this.viewMode.RECOVER;
			updateView();
		};
		menuItem.always_show_image = true;
        menuItem.show();
        mainMenu.append (menuItem);
	
        menuItem = new ImageMenuItem.with_label (_("About"));
        menuItem.image = new Gtk.Image.from_stock(Gtk.STOCK_ABOUT, Gtk.IconSize.MENU);
        menuItem.activate.connect( showAbout );
		menuItem.always_show_image = true;
        menuItem.show();
        mainMenu.append (menuItem);
		return mainMenu;
    }



    private void switchView () {
		this.preferences.view = this.currentViewMode;

        Gtk.Widget oldBox;
		if ((oldBox = this.contentBox.get_child()) != null) {
			this.contentBox.remove(oldBox);
		}

		Gtk.Image iconImage;
		Gtk.Widget content;

        /* Order for main window title.
         * First is the application name ex: Lucruri
         * Second is the view name ex: Tasks, Calendar, Tasks
         * It will look something like: "Lucruri - Notes"
         */
        string title_order = _("%1$s - %2$s");
        /* Application name */
        string title_lucruri = _("Lucruri");
        /* Window title for home/overview.*/
        string title_overview = _("Overview");
        /* Window title for tasks view.*/
        string title_tasks = _("Tasks");
        /* Window title for notes view.*/
        string title_notes = _("Notes");
        /* Window title for projects view.*/
        string title_projects = _("Projects");
        /* Window title for tickler view.*/
        string title_tickler = _("Tickler");
        /* Window title for calendar view.*/
        string title_calendar = _("Calendar");
        /* Window title for recover view.*/
        string title_revover = _("Recover notes and tasks");

        switch ( this.currentViewMode ) {
            case ( this.viewMode.CALENDAR ) : 
		        this.title = title_order.printf(title_lucruri, title_calendar);
				content = getCalendarView ();
				//FIXME check for memory leaks here
	    		iconImage =	new Gtk.Image.from_pixbuf(
					Utils.UI.getPixbuf("/icons/22x22/apps/office-calendar.png"));
				break;
            case ( this.viewMode.TICKLER ) : 
		        this.title = title_order.printf(title_lucruri, title_tickler);
				content = getTicklerView (); 
				iconImage =	new Gtk.Image.from_pixbuf(
					Utils.UI.getPixbuf("/icons/22x22/status/folder-visiting.png"));
				break;
            case ( this.viewMode.CATEGORIES ) : 
		        this.title = title_order.printf(title_lucruri, title_tasks);
				content = getCategoriesView (); 
				iconImage = new Gtk.Image.from_pixbuf(
					Utils.UI.getPixbuf("/icons/22x22/actions/bookmark-new.png"));
				break;
            case ( this.viewMode.PROJECTS ) : 
		        this.title = title_order.printf(title_lucruri, title_projects);
				content = getProjectsView (); 
				iconImage = new Gtk.Image.from_pixbuf(
					Utils.UI.getPixbuf("/icons/22x22/categories/applications-development.png"));
				break;
            case ( this.viewMode.NOTES ) : 
		        this.title = title_order.printf(title_lucruri, title_notes);
				content = getNotesView (); 
				iconImage = new Gtk.Image.from_pixbuf(
					Utils.UI.getPixbuf("/icons/22x22/apps/accessories-text-editor.png"));
				break;
            case ( this.viewMode.RECOVER ) : 
		        this.title = title_order.printf(title_lucruri, title_revover);
				content = getRecoverView (); 
				iconImage = new Gtk.Image.from_stock(Gtk.STOCK_UNDO,Gtk.IconSize.BUTTON);
				break;
            default :
		        this.title = title_order.printf(title_lucruri, title_overview);
				content = getHomeView (); 
				iconImage =	new Gtk.Image.from_stock(Gtk.STOCK_HOME,Gtk.IconSize.BUTTON);
				break;
        }

        this.contentBox.add(content);
        this.contentBox.show_all();

		this.viewButton.set_icon_widget(iconImage);
		this.viewButton.show_all();


        switch ( this.currentViewMode ) {
            case ( this.viewMode.NOTES ) :        
				if ((this.toolbarBox.child != null) && (this.toolbarBox.child != this.noteBox))
					this.toolbarBox.remove(this.toolbarBox.child);
				if (this.toolbarBox.child == null)
					this.toolbarBox.add(this.noteBox);
				break;
            case ( this.viewMode.CATEGORIES ) :        
            case ( this.viewMode.PROJECTS ) :        
				if ((this.toolbarBox.child != null) && (this.toolbarBox.child != this.taskBox))
					this.toolbarBox.remove(this.toolbarBox.child);
				if (this.toolbarBox.child == null)
					this.toolbarBox.add(this.taskBox);
				break;

            case (this.viewMode.TICKLER): 
            case (this.viewMode.CALENDAR): 
				if (this.toolbarBox.child != null) {
					this.toolbarBox.remove(this.toolbarBox.child);
				}
				break;

            case ( this.viewMode.SEARCH ) :        
            case ( this.viewMode.RECOVER ) : 
			default:
				if ((this.toolbarBox.child != null) && (this.toolbarBox.child != this.searchBox))
					this.toolbarBox.remove(this.toolbarBox.child);
				if (this.toolbarBox.child == null) {
					this.toolbarBox.add(this.searchBox);
				}
				break;
		}
		this.toolbarBox.show_all();

    }
 
     private Gtk.Widget getHomeView ()  {
		if ( homeView == null ) {
			homeView = new HomeView(this.backend, this.editArea);
		} 
		homeView.update();
		return homeView;
    }

    private Gtk.Widget getCalendarView(){
		if ( calendarView == null ) {
			calendarView = new CalendarView(this.backend, this.editArea);
		} 
		calendarView.update();
		return calendarView;
    }

    private Gtk.Widget getTicklerView(){
		if ( ticklerView == null ) {
			ticklerView = new TicklerView(this.backend, this.editArea);
		} 
		ticklerView.update();
		return ticklerView;
    }

    
    private Gtk.Widget getProjectsView ()  {
		if ( projectsView == null ) {
			projectsView = new ProjectsView(this.backend, this.editArea);
		}
		projectsView.update();
		return projectsView;
    }



    private Gtk.Widget getCategoriesView(){
		if ( categoriesView == null ) {
			categoriesView = new CategoriesView(this.backend, this.editArea);
		}
		categoriesView.update();
		return categoriesView;
    }

    private Gtk.Widget getNotesView(){
		if ( notesView == null ) {
			notesView = new NotesView(this.backend);
		}
		notesView.update();
		return notesView;
    }


    private Gtk.Widget getRecoverView(){
		if ( recoveryView == null ) {
			recoveryView = new RecoveryView(this.backend);
		}
		recoveryView.update();
		return recoveryView;

    }

    private void addTask () {
		Utils.UI.addTask(this.backend);
	}

    private void addNote () {
		Utils.UI.addNote(this.backend);
	}
     
    private void showPreferences ( Gtk.Widget sender ) {
        preferencesDialog = new PreferencesDialog( this.preferences );
        preferencesDialog.response += applyPreferences;
        preferencesDialog.run();
        preferencesDialog.destroy();
    }
    
    private void applyPreferences ( PreferencesDialog sender, int response ) {
        this.backend.username = this.preferences.tracks_username;
        this.backend.password = this.preferences.tracks_password;
        this.backend.uri = this.preferences.tracks_uri;
        update ();
    }
    
    private void showAbout ( Gtk.Widget sender) {
        AboutDialog about = new AboutDialog();
        about.show();
    }

} // end of MainWindow class
