public class NoteWindow : Gtk.Window {

    private INote note;
    private TextEditor textView;
	private IBackend backend;
	private bool initialSticky;
	private Gtk.Menu projectsMenu;
	Gtk.ToolButton notebookButton;

	private string activeProjectID;
	private Logger log;

	public string id { get { return this.note.id;}}
	
	construct {
		log = Logger.getLogger("NoteWindow");
	}

	public NoteWindow (INote aNote){
		this.note = aNote;
		this.backend = aNote.backend;
		activeProjectID = note.projectID;
		initialSticky = note.sticky;
		//autosave at every 1 minute
		//TODO add config option
		
        GLib.Timeout.add_seconds(1 * 60 , () => {
			if ((this as NoteWindow) == null) return false; 
			this.updateNote(); 
			this.title=this.note.name; 
			return true;
			}
		);

        this.title = this.note.name;
        this.icon = Utils.UI.getPixbuf ("/icons/16x16/actions/note-edit.png" );

		if ( this.note.width < 50 )	this.note.width = 420;
		if ( this.note.height < 50 ) this.note.height = 250;
		this.default_width = this.note.width;
		this.default_height = this.note.height;

		/* text view widget */ 
        textView = new TextEditor(this.note);
		/* add note specific buttons to toolbar */
		populateToolbar();
		this.add(textView);

		this.delete_event += () => {
			updateNote(); 
			Configuration.openNoteWindows.remove(this);
			this.destroy();
		};
    }

	private void updateNote () {

		bool needs_update = false;

        int current_width;
        int current_height;
        this.get_size(out current_width, out current_height);
        if ((this.note.width != current_width) ||
            (this.note.height != current_height)) {
            this.note.width = current_width;
            this.note.height = current_height;
            needs_update = true;
        }
		
		if (this.note.projectID != activeProjectID ) {
			this.note.projectID = activeProjectID;
			needs_update = true;
		}

		if (this.note.name != this.textView.title) {
			if (this.textView.title != "") {
				this.note.name = this.textView.title;
			} else {
				this.note.name = _("Untitled Note");
			}
			needs_update = true;
		}
		if (this.note.text != this.textView.text) {
			this.note.text = this.textView.text;
			needs_update = true;
		}
		if (this.note.sticky != initialSticky ) {
			needs_update = true;
		}

		if (needs_update && !this.note.isDeleted) this.note.update();
	}

	private void showProjectsMenu(Gtk.Widget parent) {
		projectsMenu = new Gtk.Menu();
		projectsMenu.attach_to_widget(parent, () => {});

		unowned SList<Gtk.RadioMenuItem>group = null;
		Gtk.RadioMenuItem radio = new Gtk.RadioMenuItem.with_label(group,"");
		group = radio.get_group();

		GenericMenuItem item;
		/* Menu entry used when item has no associated project.*/
	    item = new GenericMenuItem("0", _("_No project") );
		if (activeProjectID == "0") item.active = true;
		item.activate += setProject;
	    projectsMenu.append (item);
   		Gtk.SeparatorMenuItem separator = new Gtk.SeparatorMenuItem();
        projectsMenu.append (separator);
		foreach (IProject project in this.backend.projects) {
			if (!project.isDeleted) {
				item = new GenericMenuItem(project.id, project.name);
				item.activate += setProject;
				if (activeProjectID == project.id) item.active = true;
				projectsMenu.append (item);
			}			
		}
        projectsMenu.show_all();
        projectsMenu.popup (null, null, Utils.UI.getMenuPosition, 1, Gtk.get_current_event_time ());
	}

	private void setProject ( GenericMenuItem sender) {
		if (sender.id != this.activeProjectID) {
			activeProjectID = sender.id;
			notebookButton.set_label( (sender.id == "0") ? _("Project"): Utils.UI.ellipsify(sender.name,10));
		}
	}

    private void toggleDelete ( Gtk.ToolButton button ) {
		if (this.note.isDeleted) {
			this.note.isDeleted = false;
			this.note.created -= Utils.UI.noteCreatedCB;
			this.note.push();
			button.set_icon_widget (new Gtk.Image.from_stock(Gtk.STOCK_DELETE, Gtk.IconSize.BUTTON));
			button.set_label(_("Delete"));
			button.show_all();
		} else {
	   		Configuration.openNoteWindows.remove(this);
		    this.note.delete();
			/* call the delete_event to do the closing part*/
		    this.delete_event(new Gdk.Event(Gdk.EventType.DELETE));
		}
    }

	private void toggleSticky ( Gtk.ToolButton button ) {

		if (this.note.sticky) { 
			this.note.sticky = false;
			button.set_icon_widget (new Gtk.Image.from_pixbuf(Utils.UI.getPixbuf ("/icons/16x16/status/pin-up.png" )));
			button.set_label(_("Pin"));
		} else {
			this.note.sticky = true;
			button.set_icon_widget (new Gtk.Image.from_pixbuf( Utils.UI.getPixbuf ("/icons/16x16/status/pin-down.png" )));	
			button.set_label(_("Unpin"));
		}
		button.show_all();

	}

	private void populateToolbar() {

		Gtk.ToolButton stickyButton = (this.note.sticky) ? 
							new Gtk.ToolButton( new Gtk.Image.from_pixbuf(Utils.UI.getPixbuf ("/icons/16x16/status/pin-down.png" )), _("Unpin")) :
							new Gtk.ToolButton( new Gtk.Image.from_pixbuf(Utils.UI.getPixbuf ("/icons/16x16/status/pin-up.png" )), _("Pin"));
		stickyButton.clicked += toggleSticky;
		stickyButton.has_tooltip = true;
		stickyButton.query_tooltip += (sender, x, y, keyboard_tooltip, tooltip) => {
			(tooltip as Gtk.Tooltip).set_text(_("Make note always visible"));
			return true;
		};
		textView.toolbar.insert(stickyButton, 0);

        notebookButton = new Gtk.ToolButton( new Gtk.Image.from_pixbuf( Utils.UI.getPixbuf ("/icons/22x22/actions/note-full.png" )),_("Project"));
		/* if note is associated to a project, set the project name*/
	    /* FIXME use a better contant */
		if (note.projectID != "0") {
			foreach (IProject project in backend.projects) {
				if (project.id == note.projectID) {
					notebookButton.set_label(Utils.UI.ellipsify(project.name,10));
					break;
				}
			}
		}
		notebookButton.clicked += (sender) => {showProjectsMenu(sender);};
		notebookButton.has_tooltip = true;
		notebookButton.query_tooltip += (sender, x, y, keyboard_tooltip, tooltip) => {
			(tooltip as Gtk.Tooltip).set_text(_("Assign note to a project"));
			return true;
		};
		textView.toolbar.insert(new Gtk.SeparatorToolItem (), -1);
		textView.toolbar.insert(notebookButton, -1);

		Gtk.ToolButton deleteButton;
		if (this.note.isDeleted) {
	    	deleteButton = new Gtk.ToolButton(new Gtk.Image.from_stock(Gtk.STOCK_UNDO, Gtk.IconSize.BUTTON),_("Recover"));
		} else {
	        deleteButton = new Gtk.ToolButton(new Gtk.Image.from_stock(Gtk.STOCK_DELETE, Gtk.IconSize.BUTTON),_("Delete"));
		}

		deleteButton.clicked += toggleDelete;
		deleteButton.has_tooltip = true;
		deleteButton.query_tooltip += (sender, x, y, keyboard_tooltip, tooltip) => {
			if (this.note.isDeleted)
				(tooltip as Gtk.Tooltip).set_text(_("Recover this note"));
			else 
				(tooltip as Gtk.Tooltip).set_text(_("Delete this note"));
			return true;
		};

		textView.toolbar.insert(new Gtk.SeparatorToolItem (), -1);
		textView.toolbar.insert(deleteButton, -1);
        

	}
/* end of NoteWindow */
}

