using Gtk;

public class ProjectsView : Gtk.HPaned {

    private TasksTreeView tasksView;
	private ProjectsTreeView projectsView;
	private Gtk.TreeModelFilter filter;
	private Gtk.ListStore projectsModel;
	private IBackend backend;
	private Gtk.TreeIter selectedIter;

	private string searchString;
	private Logger log;
	private bool showDone;

	construct {
		searchString = "";
		log = Logger.getLogger("ProjectsView");
		showDone = false;
	}

    public ProjectsView (IBackend aBackend, Gtk.Widget anEditArea)
    {
		backend = aBackend;

		projectsView = new ProjectsTreeView (true);
		projectsView.currentFilterRule = projectsView.ALL_PROJECTS;
		projectsView.nameRenderer.edited += editProjectName;
	
		Gtk.HBox toolbar = new Gtk.HBox(true,0);
		Gtk.Button addProjectButton = new Gtk.Button.from_stock(Gtk.STOCK_ADD);
		Gtk.Button deleteProjectButton = new Gtk.Button.from_stock(Gtk.STOCK_REMOVE);
		addProjectButton.clicked += () => {
			IProject project = backend.newProject(true);
			project.created += (sender, status) => {
				if (status) {
				    update();
				}
			};
			project.push();
		};

		deleteProjectButton.clicked += () => {
			IProject project = this.projectsView.getSelectedProject();
			if (project == null) return;
			project.delete();
		};

		toolbar.add(addProjectButton);
		toolbar.add(deleteProjectButton);

		Gtk.VBox optionsBox = new Gtk.VBox(false, 0);
		Gtk.CheckButton doneOption = new Gtk.CheckButton.with_label (_("Show completed task"));
		doneOption.toggled += () => { 
			this.showDone = (this.showDone) ? false : true; 
			this.filter.refilter();
			};
		optionsBox.add( doneOption);

		Gtk.VBox projectsBox = new Gtk.VBox(false, 0);
		projectsBox.pack_start(projectsView, true, true, 0);
		projectsBox.pack_start(optionsBox, false, false, 0);
		projectsBox.pack_start(toolbar, false, false, 0);

		tasksView = new TasksTreeView (anEditArea, false);

        Adjustment hScrollAdjustment = new Adjustment (0,0,0,0,0,0);
        Adjustment vScrollAdjustment = new Adjustment (0,0,0,0,0,0);
        ScrolledWindow scrolledWindow = new ScrolledWindow (hScrollAdjustment, vScrollAdjustment);
        scrolledWindow.hscrollbar_policy = PolicyType.NEVER;
        scrolledWindow.vscrollbar_policy = PolicyType.AUTOMATIC;

        scrolledWindow.add(tasksView);

		position = 180;
		pack1 (projectsBox, false, false);
		pack2 (scrolledWindow, true, false);
    }

	public void update() {
			updateTasksModel();
			this.tasksView.set_model ( this.filter );
			projectsView.targetModel = this.filter;
			this.filter.refilter();

			updateProjectsModel();
			this.projectsView.set_model ( this.projectsModel );
			Gtk.TreePath? path = this.projectsModel.get_path (this.selectedIter);
			if (path != null) this.projectsView.set_cursor(path, null, false);
	}

	public void filterSearch(string text) {
		this.searchString = text;
		//TODO add timeout for search
		this.filter.refilter();
	}

/******************************************************************************
 * PRIVATE
 */

	private void editProjectName ( GLib.Object sender, string path, string newName) {
        Gtk.TreeIter iter;
        IProject project;
        
        Gtk.TreePath aPath = new Gtk.TreePath.from_string( path );
        if (!this.projectsModel.get_iter (out iter, aPath))
	    return; // Do nothing
				
		this.projectsModel.get (iter, 0, out project);	
        if ( project == null ) return;		
		if (project.name != newName) {
            project.name = newName;
            project.update();       
        }
    }

	private void updateTasksModel() {
        Gtk.TreeIter ti;
        Gtk.ListStore treeModel = new Gtk.ListStore (1, typeof(ITask));

        /*
         unpack the task list and store it in a tree model
        */
        foreach (ITask task in this.backend.tasks) {
            treeModel.append (out ti);
            treeModel.set (ti, 0, task);
        }

 		filter = new Gtk.TreeModelFilter (treeModel, null);
		filter.set_visible_func( filterTasksTree );
	}

	private void updateProjectsModel() {

        Gtk.TreeIter ti;
		ProjectsListStore treeModel = new ProjectsListStore();

        /*
         unpack the task list and store it in a tree model
        */
		IProject meta;
		meta = backend.newProject(false);
		meta.name = "<b>" + _("All Projects") + "</b>";
		meta.id = projectsView.ALL_PROJECTS;
        treeModel.append (out ti);
        treeModel.set (ti, 0, meta);
		selectedIter = ti;

		Utils.Backend.sortProjectsByPositionAsc(this.backend.projects);
        foreach (IProject project in this.backend.projects) {
				if (!project.isDeleted) {
			        treeModel.append (out ti);
			        treeModel.set (ti, 0, project);
				}
        }
		projectsModel = treeModel;
	}

    private bool filterTasksTree (Gtk.TreeModel model, Gtk.TreeIter iter) {
 	    ITask task;
	    model.get (iter, 0, out task);
		if (task == null) return false;

		if ((task.isDeleted)||(!this.showDone && task.done)) return false;

	    if (
				((projectsView.currentFilterRule == projectsView.ALL_PROJECTS) ||
				(task.projectID == projectsView.currentFilterRule) ) &&

				( (searchString == "") || (searchString == Utils.Constant.filterTasks) ||
			  (task.name.casefold().str(searchString.casefold()) != null) || 
			  (task.text.casefold().str(searchString.casefold()) != null) )

			) {
            return true;
        } else {
           return false;
		}
    }

}

public class ProjectsListStore : Gtk.ListStore, Gtk.TreeDragDest, Gtk.TreeDragSource {
	private Logger log;
	
	construct {
		log = Logger.getLogger("ProjectsListStore");
	}

	public ProjectsListStore () {
		set_column_types (new GLib.Type[] { typeof(IProject)});
	}

	public bool drag_data_delete (Gtk.TreePath path) {

	    Gtk.TreeIter iter;
        IProject project;
        
        if (!get_iter (out iter, path)) return false;
		get (iter, 0, out project);
		if (project == null) return false;

		bool val = base.drag_data_delete (path);

		get_iter_first(out iter);
		get (iter, 0, out project);
		int i = 1;
		while (iter_next(ref iter)) {
			get (iter, 0, out project);
			if (project.position != i ) {
				project.position = i;
				project.update();
			}
			i++;
		}
		project.backend.updated();
		return val;
	}

	public bool row_drop_possible (Gtk.TreePath dest, Gtk.SelectionData selection_data) {

	    Gtk.TreeIter iter;
        IProject project;
        
        if (!get_iter (out iter, dest)) return false;
		get (iter, 0, out project);
		if (project == null) return false;

		if (project.id == ProjectsTreeView.ALL_PROJECTS) {
			return false;
		} else {
			return true;
		}
	}

}
