public class ChestiiBackend : GLib.Object, IBackend {

    /*properties */
    public Gee.ArrayList<ITask> tasks { get; set; }
    public Gee.ArrayList<ICategory> categories { get; set; }
    public Gee.ArrayList<IProject> projects { get; set; }
    public Gee.ArrayList<INote> notes { get; set; }

    private const string _name = _("Chestii Backend");
    public  string name { get { 
        //Name of the backend
        return _name; 
        }
   }

    public bool isConfigured { get {
        if (( _username == "" ) || (_uri == ""))
            return false;
        else
            return true;
        }
    }

    private string _username;
    public string username { set { _username = value; } }
    private string _password;
    public string password { set { _password = value; } }
    private string _uri;
    public string uri { set { _uri = value; } }    

    private Configuration _preferences;
    public Configuration preferences { set { _preferences = value; } }


    /* signals */
    public signal void projectsUpdated ();
    public signal void categoriesUpdated ();
    public signal void tasksUpdated ();
    public signal void notesUpdated ();
        
    public signal void notFound ();
    public signal void accessDenied ();


	private Logger log;
	/* constructors */
	construct {
		log = Logger.getLogger("ChestiiBackend");
	}

    public ChestiiBackend() {

        session = new Soup.SessionAsync();
        session.authenticate +=  (sess, msg, auth, retrying) => {
            if (retrying) return;
            auth.authenticate( _username, _password);
        };

        session.max_conns=1;
        tasks = new Gee.ArrayList<ITask>();
        projects = new Gee.ArrayList<IProject>();
        categories = new Gee.ArrayList<ICategory>();
		notes = new Gee.ArrayList<INote>();
        _username = "";
        _password = "";
        _uri = "";
   }

    /* methods */
    public void updateProjects(){
        GLib.Timeout.add_seconds(2, refreshProjectsLoop);
    }
    public void updateCategories(){
        GLib.Timeout.add_seconds(2, refreshCategoriesLoop);
    }

    private int tasksGetTries = 0;

    public void updateTasks(){
        tasksGetTries = 0;
        GLib.Timeout.add_seconds(2, refreshTasksLoop);
    }

    private int notesGetTries = 0;
    public void updateNotes(){
        notesGetTries = 0;
        GLib.Timeout.add_seconds(2, refreshNotesLoop);
    }

    public void queueGetMessage (string uri, Soup.SessionCallback func) {
        Soup.Message message = new Soup.Message("GET", _uri + uri);
        session.queue_message (message, func);
    }

    public void queuePostMessage (string uri, string xml, Soup.SessionCallback func) {
        Soup.Message message = new Soup.Message ( "POST", _uri + uri );
        message.set_request("text/xml", Soup.MemoryUse.COPY, xml, xml.length );
        session.queue_message (message, func); 
    }

    public ITask newTask () {
		ChestiiTask task = new ChestiiTask(this);
		this.tasks.add(task);
		this.tasksUpdated();
		return task;
	}

    public INote newNote () {
		ChestiiNote note = new ChestiiNote(this);
		this.notes.add(note);
		this.notesUpdated();
		return note;
	}

    public IProject newProject (bool add) {
		ChestiiProject project = new ChestiiProject(this);
		if (add) {
			this.projects.add(project);
			this.projectsUpdated();
		}
		return project;
	}

    public ICategory newCategory (bool add) {
		ChestiiCategory category = new ChestiiCategory(this);
		if (add) {
			this.categories.add(category);
			this.categoriesUpdated();			
		}
		return category;
	}

    /***************************************************************************
     * Private stuff
     */

    private const int HTTP_MAX_RETRIES = 3;
    private const int HTTP_GET_OK = 0;

    private Soup.Session session;
    private string parserTextBuffer;
    private string parserCDataBuffer;

    private IProject parserProjectBuffer;
    private Gee.ArrayList<IProject> projectsTemp;
    private int projectsGetTries = 0;

    private ICategory parserCategoryBuffer;
    private int categoriesGetTries = 0;
    private Gee.ArrayList<ICategory> categoriesTemp;

    private ITask parserTaskBuffer;
    private Gee.ArrayList<ITask> tasksTemp;

    private INote parserNoteBuffer;
    private Gee.ArrayList<INote> notesTemp;

    private bool refreshProjectsLoop () {
        queueGetMessage ("/index.php/projects/xml/", readProjectsRefreshResponse); 
        
        if  ((projectsGetTries != HTTP_GET_OK )&&(projectsGetTries < HTTP_MAX_RETRIES )) {
            return true; 
        } else { 
            return false;
        }
    }

    private bool refreshCategoriesLoop () {
        queueGetMessage ("/index.php/categories/xml/", readCategoriesRefreshResponse); 
        
        if  ((categoriesGetTries != HTTP_GET_OK )&&(categoriesGetTries < HTTP_MAX_RETRIES )) {
            return true; 
        } else { 
            return false;
        }
    }

    private bool refreshTasksLoop () {
        queueGetMessage ("/index.php/tasks/xml", readTasksRefreshResponse); 
        
        if  ((tasksGetTries != HTTP_GET_OK )&&(tasksGetTries < HTTP_MAX_RETRIES )) {
            return true; 
        } else { 
            return false;
        }
    }

    private bool refreshNotesLoop () {
        queueGetMessage ("/index.php/notes/xml", readNotesRefreshResponse); 
        
        if  ((notesGetTries != HTTP_GET_OK )&&(notesGetTries < HTTP_MAX_RETRIES )) {
            return true; 
        } else { 
            return false;
        }
    }

    private void readProjectsRefreshResponse (Soup.Session session, Soup.Message message){

        if ( message.response_body.length == 0 ) {
            projectsGetTries = HTTP_MAX_RETRIES + 1;
            //only tasks will show de error info this.notFound();
            return;
        }
        
        projectsTemp = new Gee.ArrayList<IProject>(); 
		/* TODO parse only entities and send them to ITEM.fromXML() */
        GLib.MarkupParser markupParser = { projectParserStart,
                                           projectParserEnd,
                                           parserText, null, null };
        GLib.MarkupParseContext parser = new GLib.MarkupParseContext (markupParser, GLib.MarkupParseFlags.TREAT_CDATA_AS_TEXT, this, null);
        try {
            parser.parse(message.response_body.data, (long)message.response_body.length);
            parser.end_parse();
        } catch (GLib.MarkupError ex) {
            stdout.printf("%s\n", ex.message);
            projectsGetTries++;
            updateProjects();
            return;
        }
        projectsGetTries = HTTP_GET_OK;
        projects = projectsTemp;
        /* signal the fact that projects list is updated*/
        projectsUpdated();
    }


    private void readCategoriesRefreshResponse (Soup.Session session, Soup.Message message){
        if ( message.response_body.length == 0 ) {
            categoriesGetTries = HTTP_MAX_RETRIES + 1;
            //only tasks will show de error info this.notFound();
            return;
        }

        categoriesTemp = new Gee.ArrayList<ICategory>();
        GLib.MarkupParser markupParser = { categoryParserStart,
                                           categoryParserEnd,
                                           parserText, null, null };
        GLib.MarkupParseContext parser = new GLib.MarkupParseContext (markupParser, GLib.MarkupParseFlags.TREAT_CDATA_AS_TEXT, this, null);
        
        try {
            parser.parse(message.response_body.data, (long)message.response_body.length);
            parser.end_parse();
        } catch (GLib.MarkupError ex) {
            stdout.printf("%s\n", ex.message);
            /* if there is an error, incread error counter and try again */
            categoriesGetTries++;
            updateCategories();
            return;
        }
        categoriesGetTries = HTTP_GET_OK;
        categories = categoriesTemp;

        /* signal the fact that context list is updated*/
        categoriesUpdated();
    }


    private void readTasksRefreshResponse (Soup.Session session, Soup.Message message){


       if (!validResponse ( message )) return;

        tasksTemp = new Gee.ArrayList<ITask>();

        GLib.MarkupParser markupParser = { taskParserStart,
                                           taskParserEnd,
                                           parserText, null, null };
        GLib.MarkupParseContext parser = new GLib.MarkupParseContext (markupParser, GLib.MarkupParseFlags.TREAT_CDATA_AS_TEXT, this, null);
        try {
            parser.parse(message.response_body.data, (long)message.response_body.length);
            parser.end_parse();
        } catch (GLib.MarkupError ex) {
            stdout.printf("Error: %s\n", ex.message);
            tasksGetTries++;
            updateTasks();
            return;
        }

        tasksGetTries = HTTP_GET_OK;
        tasks = tasksTemp;
        /* signal the fact that task list is updated*/
        tasksUpdated();

     }


    private void readNotesRefreshResponse (Soup.Session session, Soup.Message message){


       if (!validResponse ( message )) return;

        notesTemp = new Gee.ArrayList<INote>();

        GLib.MarkupParser markupParser = { noteParserStart,
                                           noteParserEnd,
                                           parserText, null , null };
        GLib.MarkupParseContext parser = new GLib.MarkupParseContext (markupParser, GLib.MarkupParseFlags.TREAT_CDATA_AS_TEXT, this, null);
        try {
            parser.parse(message.response_body.data, (long)message.response_body.length);
            parser.end_parse();
        } catch (GLib.MarkupError ex) {
            stdout.printf("Error: %s\n", ex.message);
            notesGetTries++;
            updateNotes();
            return;
        }

        notesGetTries = HTTP_GET_OK;
        notes = notesTemp;
        /* signal the fact that task list is updated*/
        notesUpdated();

     }

    private bool validResponse ( Soup.Message message ) {
        if ( message.response_body.length == 0 ) {
            tasksGetTries = HTTP_MAX_RETRIES + 1;
            this.notFound();
            return false;
        }
        if ( message.status_code == 401 ) {
            tasksGetTries = HTTP_MAX_RETRIES + 1;
            this.accessDenied();
            return false;
            
        }
        return true;
    }

/*************************************************************************
 *    PARSERS
 */




    private void parserText (GLib.MarkupParseContext context, string text, ulong text_len) throws GLib.MarkupError {
		parserCDataBuffer = parserTextBuffer;
        parserTextBuffer = text;
    }

    /*
     * Project parsing function 
     */
    private void projectParserStart (GLib.MarkupParseContext context, string element_name, string[] attribute_names, string[] attribute_values) throws GLib.MarkupError {
        if ( element_name == "html") {
            throw new GLib.MarkupError.INVALID_CONTENT("Not a valid projects XML");
        } else if ( element_name == "project" ) {
            parserProjectBuffer = new ChestiiProject(this);
        } else {
           //print ("start project name:" + element_name + "\n");
        }
    }

    private void projectParserEnd (GLib.MarkupParseContext context, string element_name) throws GLib.MarkupError {
 
        if ( element_name == "project" ) {
            if (parserProjectBuffer!=null) projectsTemp.add (parserProjectBuffer);
        } else if ( element_name == "active" ) {
            parserProjectBuffer.active = (bool)parserTextBuffer.to_int();
        } else if ( element_name == "name" ) {
            parserProjectBuffer.name = parserCDataBuffer.replace("]]>","");
        } else if ( element_name == "text" ) {
            parserProjectBuffer.text = parserCDataBuffer.replace("]]>","");
        } else if ( element_name == "id" ) {
            parserProjectBuffer.id = parserTextBuffer.to_int();
        } else if ( element_name == "position" ) {
            parserProjectBuffer.position = parserTextBuffer.to_int();
        } else if ( element_name == "created-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserProjectBuffer.createdAt = tim;
        } else if ( element_name == "completed-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserProjectBuffer.completedAt = tim;
        } else if ( element_name == "updated-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserProjectBuffer.updatedAt = tim;
        } else {
            //print ("end " + element_name + "\n");
        }
    }

    /*
     * Category parsing function 
     */
    private void categoryParserStart (GLib.MarkupParseContext context, string element_name, string[] attribute_names, string[] attribute_values) throws GLib.MarkupError {
        if ( element_name == "html") {
            throw new GLib.MarkupError.INVALID_CONTENT("Not a valid category XML");
        } else if ( element_name == "category" ) {
            parserCategoryBuffer = new ChestiiCategory(this);
        } else {
            /*
             unhandled tag
            */
           //print ("start category name:" + element_name + "\n");
        }
    }

    private void categoryParserEnd (GLib.MarkupParseContext context, string element_name) throws GLib.MarkupError {

        if ( element_name == "category" ) {
            if (parserCategoryBuffer != null) categoriesTemp.add (parserCategoryBuffer);
        } else if ( element_name == "hide" ) {
            parserCategoryBuffer.hide = (bool)parserTextBuffer.to_int(); ;
        } else if ( element_name == "name" ) {
            parserCategoryBuffer.name = parserCDataBuffer.replace("]]>","");
        } else if ( element_name == "id" ) {
            parserCategoryBuffer.id = parserTextBuffer.to_int();
        } else if ( element_name == "position" ) {
            parserCategoryBuffer.position = parserTextBuffer.to_int();
       } else if ( element_name == "updated-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserCategoryBuffer.updatedAt = tim;
       } else if ( element_name == "created-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserCategoryBuffer.createdAt = tim;
        } else {
            /*
             unhandled tag
            */
            //print ("end " + element_name + "\n");
        }
    }

    /*
     * task parsing function 
     */
    private void taskParserStart (GLib.MarkupParseContext context, string element_name, string[] attribute_names, string[] attribute_values) throws GLib.MarkupError {
        if ( element_name == "html") {
            throw new GLib.MarkupError.INVALID_CONTENT("Not a valid task XML");
        } else if ( element_name == "task" ) {
            parserTaskBuffer = new ChestiiTask.empty();
			parserTaskBuffer.backend = this;
        } else {
            /*
             unhandled tag
            */
           //print ("start project name:" + element_name + "\n");
        }
    }


    private void taskParserEnd (GLib.MarkupParseContext context, string element_name) throws GLib.MarkupError {
 
        if ( element_name == "task" ) {
            if (parserTaskBuffer != null) tasksTemp.add (parserTaskBuffer);
        } else if ( element_name == "done" ) {
            parserTaskBuffer.done = (bool)parserTextBuffer.to_int();
        } else if ( element_name == "name" ) {
            parserTaskBuffer.name = parserCDataBuffer.replace("]]>","");
        } else if ( element_name == "text" ) {
            parserTaskBuffer.text = parserCDataBuffer.replace("]]>","");
        } else if ( element_name == "id" ) {
            parserTaskBuffer.id = parserTextBuffer.to_int();
        } else if ( element_name == "project-id" ) {
            parserTaskBuffer.projectID = parserTextBuffer.to_int();
        } else if ( element_name == "category-id" ) {
            parserTaskBuffer.categoryID = parserTextBuffer.to_int();
        } else if ( element_name == "priority" ) {
            parserTaskBuffer.priority = parserTextBuffer.to_int();
       } else if ( element_name == "created-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserTaskBuffer.createdAt = tim;
      } else if ( element_name == "updated-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserTaskBuffer.updatedAt = tim;
        } else if ( element_name == "completed-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserTaskBuffer.completedAt = tim;
        } else if ( element_name == "due-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserTaskBuffer.dueAt = tim;
         } else if ( element_name == "show-from" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserTaskBuffer.showFrom = tim;
       } else {
             /*
              unhandled tag
             */   
            //print ("end " + element_name + "\n");
        }
    }

	/* notes parser */

    private void noteParserStart (GLib.MarkupParseContext context, string element_name, string[] attribute_names, string[] attribute_values) throws GLib.MarkupError {
        if ( element_name == "html") {
            throw new GLib.MarkupError.INVALID_CONTENT("Not a valid note XML");
        } else if ( element_name == "note" ) {
            parserNoteBuffer = new ChestiiNote(this);
        } else {
            /*
             unhandled tag
            */
           //print ("start project name:" + element_name + "\n");
        }
    }

    private void noteParserEnd (GLib.MarkupParseContext context, string element_name) throws GLib.MarkupError {
 
        if ( element_name == "note" ) {
            if (parserNoteBuffer != null) notesTemp.add (parserNoteBuffer);
        } else if ( element_name == "name" ) {
            parserNoteBuffer.name = parserCDataBuffer.replace("]]>","");
        } else if ( element_name == "text" ) {
            parserNoteBuffer.text = parserCDataBuffer.replace("]]>","");
        } else if ( element_name == "sticky" ) {
            parserNoteBuffer.sticky = (parserTextBuffer == "1") ? true : false;
        } else if ( element_name == "id" ) {
            parserNoteBuffer.id = parserTextBuffer.to_int();
        } else if ( element_name == "project-id" ) {
            parserNoteBuffer.projectID = parserTextBuffer.to_int();
        } else if ( element_name == "x" ) {
            parserNoteBuffer.x = parserTextBuffer.to_int();
        } else if ( element_name == "y" ) {
            parserNoteBuffer.y = parserTextBuffer.to_int();
        } else if ( element_name == "width" ) {
            parserNoteBuffer.width = parserTextBuffer.to_int();
        } else if ( element_name == "height" ) {
            parserNoteBuffer.height = parserTextBuffer.to_int();
        } else if ( element_name == "created-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserNoteBuffer.createdAt = tim;
        } else if ( element_name == "updated-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserNoteBuffer.updatedAt = tim;
        } else {
            //print ("end " + element_name + "\n");
        }
    }

	/* end of notes parser */

}
