using Sqlite;

public class SQLiteBackend : GLib.Object, IBackend {

    /*properties */
    public Gee.ArrayList<ITask> tasks { get; set; }
    public Gee.ArrayList<ICategory> categories { get; set; }
    public Gee.ArrayList<IProject> projects { get; set; }
    public Gee.ArrayList<INote> notes { get; set; }

	private int updateCurrentStep;
	private const int updateFinalStep = 4;

    private const string _name = _("SQLite Backend");
    public  string name { get { 
        //Name of the backend
        return _name; 
        }
	}

    public bool isConfigured { get { if (_uri != "") return true; else return false;} }

    private string _username;
    public string username { set {} }
    private string _password;
    public string password { set {} }
    private string _uri;
    public string uri { 
		set {
			_uri = value;
			if (!FileUtils.test (_uri, FileTest.IS_REGULAR)) {
				this.notFound();
				_uri = "";
			} else {
				int rc = Database.open (_uri, out db);
				if (rc == OK) { 
					this.notFound();
				}
			}
		} 
	}

	public delegate void voidFunc ();

	private IBackend.voidFunc updateUIFunc;

	public Database db;

    /* signals */
    public signal void updated ();
    public signal void notFound ();
    public signal void accessDenied ();


	private Logger log;
	/* constructors */
	construct {
		log = Logger.getLogger("SQLiteBackend");
		_username = "";
		_password = "";
		_uri = "";
		tasks = new Gee.ArrayList<ITask>();
		notes = new Gee.ArrayList<INote>();
		projects = new Gee.ArrayList<IProject>();
		categories = new Gee.ArrayList<ICategory>();

	}

    public SQLiteBackend() {

	}

    /* methods */
    public void update(IBackend.voidFunc func){
    	updateUIFunc = func;
    	updateCurrentStep = 0;
    	updateProjects();
    	updateCategories();
    	updateTasks();
    	updateNotes();
    }
    
    public void updateProjects(){
		GLib.Idle.add(idleUpdateProjects);
    }
 
    public void updateCategories(){
		GLib.Idle.add(idleUpdateCategories);
    }


    public void updateTasks(){
		GLib.Idle.add(idleUpdateTasks);
    }

    public void updateNotes(){
		GLib.Idle.add(idleUpdateNotes);
    }

    public void updateConfig (string key, string val) {
		//FIXME add idle call
		log.debug("Update Config %s = %s".printf(key,val));
		string query = "UPDATE config SET ";
		query += "value = '%s' ".printf(val);
		query += " WHERE key='%s';".printf(key);
		int rc = this.db.exec(query);
		if (rc != Sqlite.OK) { log.debug("DB error: " +  this.db.errmsg ()); }
	}

    public string getConfig (string key) {
		string query = "SELECT value FROM config WHERE key = '%s';".printf(key);
		weak string[] result;
		string errmsg;
		int rowNumber;
		int columnNumber;
		string val = "";
		db.get_table (query, out result, out rowNumber, out columnNumber, out errmsg);
		if (rowNumber < 1 ) return "";
		val = result[columnNumber];
		log.debug("getConfig: key=%s , val=%s".printf(key,val));
		db.free_table(result);
		return val;
	}

    public ITask newTask () {
		SQLiteTask task = new SQLiteTask(this);
		this.tasks.add(task);
		this.updated();
		task.updated.connect(updateUI);				
		return task;
	}

    public INote newNote () {
		SQLiteNote note = new SQLiteNote(this);
		this.notes.add(note);
		this.updated();
		note.updated.connect(updateUI);				
		return note;
	}

    public IProject newProject (bool add) {
		SQLiteProject project = new SQLiteProject(this);
		if (add) {
			this.projects.add(project);
			this.updated();
			project.updated.connect(updateUI);				
		}
		return project;
	}

    public ICategory newCategory (bool add) {
		SQLiteCategory category = new SQLiteCategory(this);
		if (add) {
			this.categories.add(category);
			this.updated();
			category.updated.connect(updateUI);				
		}
		return category;
	}

    /***************************************************************************
     * Private stuff
     */

	private void updateUI(bool status) {
		updateUIFunc();
	}

	private bool idleUpdateProjects(){
		log.debug("Start updateProjects");
		string query = "SELECT * FROM projects ORDER BY position ASC;";
		weak string[] result;
		string errmsg;
		int rowNumber;
		int columnNumber;
		db.get_table (query, out result, out rowNumber, out columnNumber, out errmsg);
		if (rowNumber < 1 ) return false;

		HashTable<string,int> columnIndex = new HashTable<string,int>(GLib.str_hash, GLib.str_equal);
		for (int i = 0; i < columnNumber; i++) {
			columnIndex.insert(result[i],i);
		}

		for (int i = columnNumber; i < ((rowNumber+1) * columnNumber) ; i += columnNumber) {
			string[] rows = new string[columnNumber];
			for (int j=0; j < columnNumber; j++) rows[j] = result[i+j];
			SQLiteProject project = new SQLiteProject.fromArray(this, rows , columnIndex);
			if ( isNewItem (projects, project)) {
				projects.add(project);
				project.updated.connect(updateUI);
			}
		}
		columnIndex.remove_all();
		db.free_table(result);
		log.debug("End update projects");
        updateCurrentStep++;
        if (updateCurrentStep >= updateFinalStep) {
        	log.debug("Backend fully updated");
        	updated();
		}
		return false;
    }
 
	private bool idleUpdateCategories(){
		log.debug("Start update categorires");
		string query = "SELECT * FROM categories ORDER BY position ASC;";
		weak string[] result;
		string errmsg;
		int rowNumber;
		int columnNumber;
		db.get_table (query, out result, out rowNumber, out columnNumber, out errmsg);
		if (rowNumber < 1 ) return false;

		HashTable<string,int> columnIndex = new HashTable<string,int>(GLib.str_hash, GLib.str_equal);
		for (int i = 0; i < columnNumber; i++) {
			columnIndex.insert(result[i],i);
		}
		for (int i = columnNumber; i < ((rowNumber+1) * columnNumber) ; i += columnNumber) {
			string[] rows = new string[columnNumber];
			for (int j=0; j < columnNumber; j++) rows[j] = result[i+j];
			SQLiteCategory category = new SQLiteCategory.fromArray(this, rows , columnIndex);
			if ( isNewItem (categories, category)) {
				categories.add(category);
				category.updated.connect(updateUI);
			}
		}
		columnIndex.remove_all();
		db.free_table(result);
		log.debug("End update categories");
        updateCurrentStep++;
        if (updateCurrentStep >= updateFinalStep) {
        	log.debug("Backend fully updated");
        	updated();
		}
		return false;

    }


	private bool idleUpdateTasks(){
		log.debug("Start update tasks");
		string query = "SELECT * FROM tasks;";
		weak string[] result;
		string errmsg;
		int rowNumber;
		int columnNumber;
		db.get_table (query, out result, out rowNumber, out columnNumber, out errmsg);
		if (rowNumber < 1 ) return false;

		HashTable<string,int> columnIndex = new HashTable<string,int>(GLib.str_hash, GLib.str_equal);
		for (int i = 0; i < columnNumber; i++) {
			columnIndex.insert(result[i],i);
		}
		for (int i = columnNumber; i < ((rowNumber+1) * columnNumber) ; i += columnNumber) {
			string[] rows = new string[columnNumber];
			for (int j=0; j < columnNumber; j++) rows[j] = result[i+j];
			SQLiteTask task = new SQLiteTask.fromArray(this, rows , columnIndex);
			if ( isNewItem(tasks,task)) {
				tasks.add(task);
				task.updated.connect(updateUI);
			}
		}
		columnIndex.remove_all();
		db.free_table(result);
		log.debug("End update tasks");
        updateCurrentStep++;
        if (updateCurrentStep >= updateFinalStep) {
        	log.debug("Backend fully updated");
        	updated();
		}
		return false;
    }

    private bool idleUpdateNotes(){
		log.debug("Start update notes");
		string query = "SELECT * FROM notes;";
		weak string[] result;
		string errmsg;
		int rowNumber;
		int columnNumber;
		db.get_table (query, out result, out rowNumber, out columnNumber, out errmsg);
		if (rowNumber < 1 ) return false;

		HashTable<string,int> columnIndex = new HashTable<string,int>(GLib.str_hash, GLib.str_equal);
		for (int i = 0; i < columnNumber; i++) {
			columnIndex.insert(result[i],i);
		}
		for (int i = columnNumber; i < ((rowNumber+1) * columnNumber) ; i += columnNumber) {
			string[] rows = new string[columnNumber];
			for (int j=0; j < columnNumber; j++) rows[j] = result[i+j];
			SQLiteNote note = new SQLiteNote.fromArray(this, rows , columnIndex);
			if ( isNewItem (notes, note)) {  
				notes.add(note);
				note.updated.connect(updateUI);
			}
		}
		columnIndex.remove_all();
		db.free_table(result);
		log.debug("End update notes");
        updateCurrentStep++;
        if (updateCurrentStep >= updateFinalStep) {
        	log.debug("Backend fully updated");
        	updated();
		}
		return false;
    }


	private bool isNewItem(Gee.ArrayList<IItem>items, IItem item) {
		foreach (IItem aItem in items) {
			if (aItem.id == item.id) return false;
		}
		return true;
	}
}
