public class TracksBackend : GLib.Object, ITasksBackend {

    /*properties */
    public Gee.ArrayList<ITask> tasks { get; set; }
    public Gee.ArrayList<ICategory> categories { get; set; }
    public Gee.ArrayList<IProject> projects { get; set; }


    private const string _name = _("Get on Tracks Backend");
    public  string name { get { 
        //Name of the backend
        return _name; 
        }
   }

    public bool isConfigured { get {
        if (( _username == "" ) || (_uri == ""))
            return false;
        else
            return true;
        }
    }

    private string _username;
    public string username { set { _username = value; } }
    private string _password;
    public string password { set { _password = value; } }
    private string _uri;
    public string uri { set { _uri = value; } }    

    private Preferences _preferences;
    public Preferences preferences { set { _preferences = value; } }


    public TracksBackend() {

        session = new Soup.SessionAsync();
        session.max_conns=1;
        tasks = new Gee.ArrayList<ITask>();
        projects = new Gee.ArrayList<IProject>();
        categories = new Gee.ArrayList<ICategory>();
        _username = "";
        _password = "";
        _uri = "";
   }


    /* signals */
    public signal void projectsUpdated ();
    public signal void categoriesUpdated ();
    public signal void tasksUpdated ();
    public signal void taskAdded ( bool status );
    public signal void taskDeleted ( bool status );
    public signal void taskToggled ( bool status );
    public signal void taskUpdated ( bool status );
    public signal void notFound ();
    public signal void accessDenied ();


    /* methods */
    public void updateProjects(){
        GLib.Timeout.add_seconds(2, refreshProjectsLoop);
    }
    public void updateCategories(){
        GLib.Timeout.add_seconds(2, refreshCategoriesLoop);
    }

    private int tasksGetTries = 0;
    public void updateTasks(){
        tasksGetTries = 0;
        GLib.Timeout.add_seconds(2, refreshTasksLoop);
    }
    
    public void addTask ( string name, int projectID, int categoryID ) {
        TracksTask task = new TracksTask();
        task.backend = this;
        task.name = name;
        task.categoryID = categoryID;
        task.projectID = projectID;
        queuePostMessage ( _uri + "/todos.xml", task.to_xml(), readAddResponse);
    }

    public void deleteTask ( ITask task ) {
        string uri = _uri + "/todos/%d.xml".printf(task.id);
        queueDeleteMessage ( uri, readDeleteResponse);
    }

    public void toggleTask ( ITask task ) {
        string xml = "";
        string uri = _uri + "/todos/%d/toggle_check.xml".printf(task.id);
        queuePutMessage ( uri, xml, readToggleResponse);   
        // TODO sync model with web
        task.toggle();

     }
    
    public void updateTask ( ITask task ) {
        string xml = (task as TracksTask).to_xml();
        string uri = _uri + "/todos/%d.xml".printf(task.id);
        stdout.printf("To web:\n %s\n",xml); stdout.flush();
        queuePutMessage ( uri, xml, readUpdateResponse);
    }

    /***************************************************************************
     * Private stuff
     */

    private const int HTTP_MAX_RETRIES = 3;
    private const int HTTP_GET_OK = 0;

    private Soup.Session session;
    private string parserTextBuffer;

    private IProject parserProjectBuffer;
    private Gee.ArrayList<IProject> projectsTemp;
    private int projectsGetTries = 0;

    private ICategory parserCategoryBuffer;
    private int categoriesGetTries = 0;
    private Gee.ArrayList<ICategory> categoriesTemp;

    private ITask parserTaskBuffer;
    private Gee.ArrayList<ITask> tasksTemp;

    private bool refreshProjectsLoop () {
        queueGetMessage (_uri + "/projects.xml", readProjectsRefreshResponse); 
        
        if  ((projectsGetTries != HTTP_GET_OK )&&(projectsGetTries < HTTP_MAX_RETRIES )) {
            return true; 
        } else { 
            return false;
        }
    }

    private bool refreshCategoriesLoop () {
        queueGetMessage (_uri + "/contexts.xml", readCategoriesRefreshResponse); 
        
        if  ((categoriesGetTries != HTTP_GET_OK )&&(categoriesGetTries < HTTP_MAX_RETRIES )) {
            return true; 
        } else { 
            return false;
        }
    }

    private bool refreshTasksLoop () {
        queueGetMessage (_uri + "/todos.xml", readTasksRefreshResponse); 
        
        if  ((tasksGetTries != HTTP_GET_OK )&&(tasksGetTries < HTTP_MAX_RETRIES )) {
            return true; 
        } else { 
            return false;
        }
    }

    private void queueGetMessage (string uri, Soup.SessionCallback func) {
        Soup.Message message = new Soup.Message("GET",uri);
        session.authenticate +=  (sess, msg, auth, retrying) => {
            if (retrying) return;
            auth.authenticate( _username, _password);
        };
        session.queue_message (message, func);
    }


    private void queueDeleteMessage (string uri, Soup.SessionCallback func) {
        Soup.Message message = new Soup.Message ( "DELETE", uri );
        session.authenticate +=  (sess, msg, auth, retrying) => {
            if (retrying) return;
            auth.authenticate( _username, _password);
        };
        session.queue_message (message, func); 
    }

    private void queuePostMessage (string uri, string xml, Soup.SessionCallback func) {
        Soup.Message message = new Soup.Message ( "POST", uri );
        message.set_request("text/xml", Soup.MemoryUse.COPY, xml, xml.length );
        session.authenticate +=  (sess, msg, auth, retrying) => {
            if (retrying) return;
            auth.authenticate( _username, _password);
        };
        session.queue_message (message, func); 
    }


    private void queuePutMessage (string uri, string xml, Soup.SessionCallback func) {
        Soup.Message message = new Soup.Message ( "PUT", uri );
        message.set_request("text/xml", Soup.MemoryUse.COPY, xml, xml.length );
        session.authenticate +=  (sess, msg, auth, retrying) => {
            if (retrying) return;
            auth.authenticate( _username, _password);
        };
        session.queue_message (message, func); 
    }

    private void readAddResponse (Soup.Session session, Soup.Message message){
        bool status;
        status = ( message.response_body.length > 1 ) ? false : true ;
        stdout.printf("Cont: %s\n", message.response_body.data);
        if (status) {
            message.response_headers.foreach( (name, val) => {  
                stdout.printf("Name: %s -> Value: %s\n", name, val);
            });
            // FIXME parse location and request only this task 
        }
        updateTasks();
        taskAdded (status);
        stdout.flush();
    }


   private void readDeleteResponse (Soup.Session session, Soup.Message message){
        bool status = true;
        //status = ( message.response_body.length > 1 ) ? false : true ;
        stdout.printf("Cont: %s\n", message.response_body.data);
        if (status) {
        message.response_headers.foreach( (name, val) => {  
                stdout.printf("Name: %s -> Value: %s\n", name, val);
            });
        }
        stdout.flush();
        taskDeleted (status);
    }

    private void readToggleResponse (Soup.Session session, Soup.Message message){
        bool status = true;
        //status = ( message.response_body.length > 1 ) ? false : true ;
        stdout.printf("Cont: %s\n", message.response_body.data);
        if (status) {
        message.response_headers.foreach( (name, val) => {  
                stdout.printf("Name: %s -> Value: %s\n", name, val);
            });
        }
        stdout.flush();
        taskToggled (status);
    }


   private void readUpdateResponse (Soup.Session session, Soup.Message message){
        bool status = true;
        //status = ( message.response_body.length > 1 ) ? false : true ;
        stdout.printf("Cont: %s\n", message.response_body.data);
        if (status) {
        message.response_headers.foreach( (name, val) => {  
                stdout.printf("Name: %s -> Value: %s\n", name, val);
            });
        }
        stdout.flush();
        taskUpdated (status);
    }


    private void readProjectsRefreshResponse (Soup.Session session, Soup.Message message){

        if ( message.response_body.length == 0 ) {
            projectsGetTries = HTTP_MAX_RETRIES + 1;
            //only tasks will show de error info this.notFound();
            return;
        }
        
        projectsTemp = new Gee.ArrayList<IProject>(); 

        GLib.MarkupParser markupParser = { projectParserStart,
                                           projectParserEnd,
                                           parserText, null, null };
        GLib.MarkupParseContext parser = new GLib.MarkupParseContext (markupParser, GLib.MarkupParseFlags.TREAT_CDATA_AS_TEXT, this, null);
        try {
            parser.parse(message.response_body.data, (long)message.response_body.length);
            parser.end_parse();
        } catch (GLib.MarkupError ex) {
            stdout.printf("%s\n", ex.message);
            projectsGetTries++;
            updateProjects();
            return;
        }
        projectsGetTries = HTTP_GET_OK;
        projects = projectsTemp;
        /* signal the fact that projects list is updated*/
        projectsUpdated();
    }


    private void readCategoriesRefreshResponse (Soup.Session session, Soup.Message message){
        if ( message.response_body.length == 0 ) {
            categoriesGetTries = HTTP_MAX_RETRIES + 1;
            //only tasks will show de error info this.notFound();
            return;
        }

        categoriesTemp = new Gee.ArrayList<ICategory>();
        GLib.MarkupParser markupParser = { categoryParserStart,
                                           categoryParserEnd,
                                           parserText, null, null };
        GLib.MarkupParseContext parser = new GLib.MarkupParseContext (markupParser, GLib.MarkupParseFlags.TREAT_CDATA_AS_TEXT, this, null);
        
        try {
            parser.parse(message.response_body.data, (long)message.response_body.length);
            parser.end_parse();
        } catch (GLib.MarkupError ex) {
            stdout.printf("%s\n", ex.message);
            /* if there is an error, incread error counter and try again */
            categoriesGetTries++;
            updateCategories();
            return;
        }
        categoriesGetTries = HTTP_GET_OK;
        categories = categoriesTemp;

        /* signal the fact that context list is updated*/
        categoriesUpdated();
    }


    private void readTasksRefreshResponse (Soup.Session session, Soup.Message message){


       if (!validResponse ( message )) return;

        tasksTemp = new Gee.ArrayList<ITask>();

        GLib.MarkupParser markupParser = { taskParserStart,
                                           taskParserEnd,
                                           parserText, null, null };
        GLib.MarkupParseContext parser = new GLib.MarkupParseContext (markupParser, GLib.MarkupParseFlags.TREAT_CDATA_AS_TEXT, this, null);
        try {
            parser.parse(message.response_body.data, (long)message.response_body.length);
            parser.end_parse();
        } catch (GLib.MarkupError ex) {
            stdout.printf("Error: %s\n", ex.message);
            tasksGetTries++;
            updateTasks();
            return;
        }

        tasksGetTries = HTTP_GET_OK;
        tasks = tasksTemp;
        /* signal the fact that task list is updated*/
        tasksUpdated();

     }

    private bool validResponse ( Soup.Message message ) {
        
       stdout.printf("CODE: %u\n", message.status_code);
    
        if ( message.response_body.length == 0 ) {
            tasksGetTries = HTTP_MAX_RETRIES + 1;
            this.notFound();
            return false;
        }
        if ( message.status_code == 401 ) {
            tasksGetTries = HTTP_MAX_RETRIES + 1;
            this.accessDenied();
            return false;
            
        }
        return true;
    }

    /*------------------------------------------------------------------
     PARSERS
    */




    private void parserText (GLib.MarkupParseContext context, string text, ulong text_len) throws GLib.MarkupError {
        //stdout.printf ("text %ld|%s|\n",(long)text_len,text);
        parserTextBuffer = text;
    }


    /*
     * Project parsing function 
     */
    private void projectParserStart (GLib.MarkupParseContext context, string element_name, string[] attribute_names, string[] attribute_values) throws GLib.MarkupError {
        if ( element_name == "html") {
            throw new GLib.MarkupError.INVALID_CONTENT("Not a valid projects XML");
        } else if ( element_name == "project" ) {
            parserProjectBuffer = new TracksProject();
        } else {
           //print ("start project name:" + element_name + "\n");
        }
    }

    private void projectParserEnd (GLib.MarkupParseContext context, string element_name) throws GLib.MarkupError {
 
        if ( element_name == "project" ) {
            if (parserProjectBuffer!=null) projectsTemp.add (parserProjectBuffer);
        } else if ( element_name == "state" ) {
            parserProjectBuffer.active = ( parserTextBuffer == "active" ) ? true : false ;
        } else if ( element_name == "name" ) {
            parserProjectBuffer.name = parserTextBuffer;
        } else if ( element_name == "description" ) {
            parserProjectBuffer.description = parserTextBuffer;
        } else if ( element_name == "id" ) {
            parserProjectBuffer.id = parserTextBuffer.to_int();
        } else if ( element_name == "position" ) {
            parserProjectBuffer.position = parserTextBuffer.to_int();
        } else if ( element_name == "default-context-id" ) {
            parserProjectBuffer.defaultContextID = parserTextBuffer.to_int();
        } else if ( element_name == "created-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserProjectBuffer.createdAt = tim;
        } else if ( element_name == "completed-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserProjectBuffer.completedAt = tim;
        } else if ( element_name == "updated-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserProjectBuffer.updatedAt = tim;
        } else {
            //print ("end " + element_name + "\n");
        }
    }

    /*
     * Category parsing function 
     */
    private void categoryParserStart (GLib.MarkupParseContext context, string element_name, string[] attribute_names, string[] attribute_values) throws GLib.MarkupError {
        if ( element_name == "html") {
            throw new GLib.MarkupError.INVALID_CONTENT("Not a valid category XML");
        } else if ( element_name == "context" ) {
            parserCategoryBuffer = new TracksCategory();
        } else {
            /*
             unhandled tag
            */
           //print ("start category name:" + element_name + "\n");
        }
    }


    private void categoryParserEnd (GLib.MarkupParseContext context, string element_name) throws GLib.MarkupError {
 
        if ( element_name == "context" ) {
            categoriesTemp.add (parserCategoryBuffer);
        } else if ( element_name == "hide" ) {
            parserCategoryBuffer.hide = ( parserTextBuffer == "true" ) ? true : false ;
        } else if ( element_name == "name" ) {
            parserCategoryBuffer.name = parserTextBuffer;
        } else if ( element_name == "id" ) {
            parserCategoryBuffer.id = parserTextBuffer.to_int();
        } else if ( element_name == "position" ) {
            parserCategoryBuffer.position = parserTextBuffer.to_int();
       } else if ( element_name == "updated-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserCategoryBuffer.updatedAt = tim;
       } else if ( element_name == "created-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserCategoryBuffer.createdAt = tim;
        } else {
            /*
             unhandled tag
            */
            //print ("end " + element_name + "\n");
        }
    }

    /*
     * task parsing function 
     */
    private void taskParserStart (GLib.MarkupParseContext context, string element_name, string[] attribute_names, string[] attribute_values) throws GLib.MarkupError {
        if ( element_name == "html") {
            throw new GLib.MarkupError.INVALID_CONTENT("Not a valid task XML");
        } else if ( element_name == "todo" ) {
            parserTaskBuffer = new TracksTask();
            parserTaskBuffer.backend = this;
        } else {
            /*
             unhandled tag
            */
           //print ("start project name:" + element_name + "\n");
        }
    }


    private void taskParserEnd (GLib.MarkupParseContext context, string element_name) throws GLib.MarkupError {
 
        if ( element_name == "todo" ) {
            if (parserTaskBuffer!=null) tasksTemp.add (parserTaskBuffer);
        } else if ( element_name == "state" ) {
            parserTaskBuffer.done = ( parserTextBuffer == "completed" ) ? true : false ;
        } else if ( element_name == "description" ) {
            parserTaskBuffer.name = parserTextBuffer;
        } else if ( element_name == "notes" ) {
            parserTaskBuffer.text = parserTextBuffer;
        } else if ( element_name == "id" ) {
            parserTaskBuffer.id = parserTextBuffer.to_int();
        } else if ( element_name == "project-id" ) {
            parserTaskBuffer.projectID = parserTextBuffer.to_int();
        } else if ( element_name == "context-id" ) {
            parserTaskBuffer.categoryID = parserTextBuffer.to_int();
        } else if ( element_name == "priority" ) {
            parserTaskBuffer.priority = parserTextBuffer.to_int();
       } else if ( element_name == "created-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserTaskBuffer.createdAt = tim;
      } else if ( element_name == "updated-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserTaskBuffer.updatedAt = tim;
        } else if ( element_name == "completed-at" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserTaskBuffer.completedAt = tim;
        } else if ( element_name == "due" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserTaskBuffer.dueAt = tim;
         } else if ( element_name == "show-from" ) {
            GLib.Time tim = GLib.Time();
            tim.strptime (parserTextBuffer, "%Y-%m-%dT%T");
            parserTaskBuffer.showFrom = tim;
       } else {
             /*
              unhandled tag
             */   
            //print ("end " + element_name + "\n");
        }
    }


}
